"""
Tests for the udiskie.cache module.
"""

import unittest
import time

from udiskie.cache import PasswordCache


class TestDev:

    def __init__(self, id_uuid):
        self.id_uuid = id_uuid


class TestPasswordCache(unittest.TestCase):

    """
    Tests for the udiskie.cache.PasswordCache class.
    """

    # NOTE: The device names are different in each test so that they do not
    # interfere accidentally.

    def test_timeout(self):
        """The cached password expires after the specified timeout."""
        device = TestDev('ALPHA')
        password = '{<}hëllo ωορλδ!{>}'
        cache = PasswordCache(1)
        cache[device] = password
        self.assertEqual(cache[device], password.encode('utf-8'))
        time.sleep(1.5)
        with self.assertRaises(KeyError):
            cache[device]

    def test_touch(self):
        """Key access refreshes the timeout."""
        device = TestDev('BETA')
        password = '{<}hëllo ωορλδ!{>}'
        cache = PasswordCache(3)
        cache[device] = password
        time.sleep(2)
        self.assertEqual(cache[device], password.encode('utf-8'))
        time.sleep(2)
        self.assertEqual(cache[device], password.encode('utf-8'))
        time.sleep(4)
        with self.assertRaises(KeyError):
            cache[device]

    def test_revoke(self):
        """A key can be deleted manually."""
        device = TestDev('GAMMA')
        password = '{<}hëllo ωορλδ!{>}'
        cache = PasswordCache(0)
        cache[device] = password
        self.assertEqual(cache[device], password.encode('utf-8'))
        del cache[device]
        with self.assertRaises(KeyError):
            cache[device]

    def test_update(self):
        device = TestDev('DELTA')
        password = '{<}hëllo ωορλδ!{>}'
        cache = PasswordCache(0)
        cache[device] = password
        self.assertEqual(cache[device], password.encode('utf-8'))
        cache[device] = password * 2
        self.assertEqual(cache[device], password.encode('utf-8')*2)
        del cache[device]
        with self.assertRaises(KeyError):
            cache[device]


if __name__ == '__main__':
    unittest.main()



"""
Common DBus utilities.
"""

from functools import partial

from gi.repository import Gio
from gi.repository import GLib

from .async_ import gio_callback, pack, Future


__all__ = [
    'InterfaceProxy',
    'PropertiesProxy',
    'ObjectProxy',
    'BusProxy',
    'connect_service',
    'MethodsProxy',
]


unpack_variant = GLib.Variant.unpack


async def call(proxy, method_name, signature, args, flags=0, timeout_msec=-1):
    """
    Asynchronously call the specified method on a DBus proxy object.

    :param Gio.DBusProxy proxy:
    :param str method_name:
    :param str signature:
    :param tuple args:
    :param int flags:
    :param int timeout_msec:
    """
    future = Future()
    cancellable = None
    proxy.call(
        method_name,
        GLib.Variant(signature, tuple(args)),
        flags,
        timeout_msec,
        cancellable,
        gio_callback,
        future,
    )
    result = await future
    value = proxy.call_finish(result)
    return pack(*unpack_variant(value))


async def call_with_fd_list(proxy, method_name, signature, args, fds,
                            flags=0, timeout_msec=-1):
    """
    Asynchronously call the specified method on a DBus proxy object.

    :param Gio.DBusProxy proxy:
    :param str method_name:
    :param str signature:
    :param tuple args:
    :param list fds:
    :param int flags:
    :param int timeout_msec:
    """
    future = Future()
    cancellable = None
    fd_list = Gio.UnixFDList.new_from_array(fds)
    proxy.call_with_unix_fd_list(
        method_name,
        GLib.Variant(signature, tuple(args)),
        flags,
        timeout_msec,
        fd_list,
        cancellable,
        gio_callback,
        future,
    )
    result = await future
    value, fds = proxy.call_with_unix_fd_list_finish(result)
    return pack(*unpack_variant(value))


class InterfaceProxy:

    """
    DBus proxy object for a specific interface.

    :ivar str object_path: object path of the DBus object
    :ivar Gio.DBusProxy _proxy: underlying proxy object
    """

    def __init__(self, proxy):
        """
        Initialize property and method attribute accessors for the interface.

        :param Gio.DBusProxy proxy: accessed object
        """
        self._proxy = proxy
        self.object_path = proxy.get_object_path()

    @property
    def object(self):
        """Get an ObjectProxy instanec for the underlying object."""
        proxy = self._proxy
        return ObjectProxy(proxy.get_connection(),
                           proxy.get_name(),
                           proxy.get_object_path())

    def connect(self, event, handler):
        """Connect to a DBus signal, returns subscription id (int)."""
        interface = self._proxy.get_interface_name()
        return self.object.connect(interface, event, handler)

    def call(self, method_name, signature='()', *args):
        return call(self._proxy, method_name, signature, args)


class PropertiesProxy(InterfaceProxy):

    Interface = 'org.freedesktop.DBus.Properties'

    def __init__(self, proxy, interface_name=None):
        super().__init__(proxy)
        self.interface_name = interface_name

    def GetAll(self, interface_name=None):
        return self.call('GetAll', '(s)',
                         interface_name or self.interface_name)


class ObjectProxy:

    """Simple proxy class for a DBus object."""

    def __init__(self, connection, bus_name, object_path):
        """
        Initialize member variables.

        :ivar Gio.DBusConnection connection:
        :ivar str bus_name:
        :ivar str object_path:

        This performs no IO at all.
        """
        self.connection = connection
        self.bus_name = bus_name
        self.object_path = object_path

    def _get_interface(self, name):
        """Get a Future(Gio.DBusProxy) for the specified interface."""
        return proxy_new(
            self.connection,
            Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES |
            Gio.DBusProxyFlags.DO_NOT_CONNECT_SIGNALS,
            info=None,
            name=self.bus_name,
            object_path=self.object_path,
            interface_name=name,
        )

    async def get_interface(self, name):
        """Get an InterfaceProxy for the specified interface."""
        proxy = await self._get_interface(name)
        return InterfaceProxy(proxy)

    async def get_property_interface(self, interface_name=None):
        proxy = await self._get_interface(PropertiesProxy.Interface)
        return PropertiesProxy(proxy, interface_name)

    @property
    def bus(self):
        """Get a BusProxy for the underlying bus."""
        return BusProxy(self.connection, self.bus_name)

    def connect(self, interface, event, handler):
        """Connect to a DBus signal. Returns subscription id (int)."""
        object_path = self.object_path
        return self.bus.connect(interface, event, object_path, handler)

    async def call(self, interface_name, method_name, signature='()', *args):
        proxy = await self.get_interface(interface_name)
        result = await proxy.call(method_name, signature, *args)
        return result


class BusProxy:

    """
    Simple proxy class for a connected bus.

    :ivar Gio.DBusConnection connection:
    :ivar str bus_name:
    """

    def __init__(self, connection, bus_name):
        """
        Initialize member variables.

        :param Gio.DBusConnection connection:
        :param str bus_name:

        This performs IO at all.
        """
        self.connection = connection
        self.bus_name = bus_name

    def get_object(self, object_path):
        """Get an ObjectProxy representing the specified object."""
        return ObjectProxy(self.connection, self.bus_name, object_path)

    def connect(self, interface, event, object_path, handler):
        """
        Connect to a DBus signal. If ``object_path`` is None, subscribe for
        all objects and invoke the callback with the object_path as its first
        argument.
        """
        if object_path:
            def callback(connection, sender_name, object_path,
                         interface_name, signal_name, parameters):
                return handler(*unpack_variant(parameters))
        else:
            def callback(connection, sender_name, object_path,
                         interface_name, signal_name, parameters):
                return handler(object_path, *unpack_variant(parameters))
        return self.connection.signal_subscribe(
            self.bus_name,
            interface,
            event,
            object_path,
            None,
            Gio.DBusSignalFlags.NONE,
            callback,
        )

    def disconnect(self, subscription_id):
        """Disconnect a DBus signal subscription."""
        self.connection.signal_unsubscribe(subscription_id)


async def proxy_new(connection, flags, info, name, object_path, interface_name):
    """Asynchronously call the specified method on a DBus proxy object."""
    future = Future()
    cancellable = None
    Gio.DBusProxy.new(
        connection,
        flags,
        info,
        name,
        object_path,
        interface_name,
        cancellable,
        gio_callback,
        future,
    )
    result = await future
    value = Gio.DBusProxy.new_finish(result)
    if value is None:
        raise RuntimeError("Failed to connect DBus object!")
    return value


async def proxy_new_for_bus(bus_type, flags, info, name, object_path,
                            interface_name):
    """Asynchronously call the specified method on a DBus proxy object."""
    future = Future()
    cancellable = None
    Gio.DBusProxy.new_for_bus(
        bus_type,
        flags,
        info,
        name,
        object_path,
        interface_name,
        cancellable,
        gio_callback,
        future,
    )
    result = await future
    value = Gio.DBusProxy.new_for_bus_finish(result)
    if value is None:
        raise RuntimeError("Failed to connect DBus object!")
    return value


async def connect_service(bus_name, object_path, interface):
    """Connect to the service object on DBus, return InterfaceProxy."""
    proxy = await proxy_new_for_bus(
        Gio.BusType.SYSTEM,
        Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES |
        Gio.DBusProxyFlags.DO_NOT_CONNECT_SIGNALS,
        info=None,
        name=bus_name,
        object_path=object_path,
        interface_name=interface,
    )
    return InterfaceProxy(proxy)


class MethodsProxy:

    """Provide methods as attributes for one interface of a DBus object."""

    def __init__(self, object_proxy, interface_name):
        """Initialize from (ObjectProxy, str)."""
        self._object_proxy = object_proxy
        self._interface_name = interface_name

    def __getattr__(self, name):
        """Get a proxy for the specified method on this interface."""
        return partial(self._object_proxy.call, self._interface_name, name)



"""
User prompt utility.
"""

from udiskie.depend import has_Gtk, require_Gtk
from udiskie.common import is_utf8

from distutils.spawn import find_executable
import getpass
import logging
import shlex
import string
import subprocess
import sys

try:
    from importlib.resources import read_text
except ImportError:  # for Python<3.7
    from importlib_resources import read_text

from .async_ import exec_subprocess, run_bg, Future
from .locale import _
from .config import DeviceFilter

Gtk = None

__all__ = ['password', 'browser']


dialog_definition = read_text(__package__, 'password_dialog.ui')


class Dialog(Future):

    def __init__(self, window):
        self._enter_count = 0
        self.window = window
        self.window.connect("response", self._result_handler)

    def _result_handler(self, window, response):
        self.set_result(response)

    def __enter__(self):
        self._enter_count += 1
        self._awaken()
        return self

    def __exit__(self, *exc_info):
        self._enter_count -= 1
        if self._enter_count == 0:
            self._cleanup()

    def _awaken(self):
        self.window.present()

    def _cleanup(self):
        self.window.hide()
        self.window.destroy()


class PasswordResult:
    def __init__(self, password=None, cache_hint=None):
        self.password = password
        self.cache_hint = cache_hint


class PasswordDialog(Dialog):

    INSTANCES = {}
    content = None

    @classmethod
    def create(cls, key, title, message, options):
        if key in cls.INSTANCES:
            return cls.INSTANCES[key]
        return cls(key, title, message, options)

    def _awaken(self):
        self.INSTANCES[self.key] = self
        super()._awaken()

    def _cleanup(self):
        del self.INSTANCES[self.key]
        super()._cleanup()

    def __init__(self, key, title, message, options):
        self.key = key
        global Gtk
        Gtk = require_Gtk()
        builder = Gtk.Builder.new()
        builder.add_from_string(dialog_definition)
        window = builder.get_object('entry_dialog')
        self.entry = builder.get_object('entry')

        show_password = builder.get_object('show_password')
        show_password.set_label(_('Show password'))
        show_password.connect('clicked', self.on_show_password)

        allow_keyfile = options.get('allow_keyfile')
        keyfile_button = builder.get_object('keyfile_button')
        keyfile_button.set_label(_('Open keyfile…'))
        keyfile_button.set_visible(allow_keyfile)
        keyfile_button.connect('clicked', run_bg(self.on_open_keyfile))

        allow_cache = options.get('allow_cache')
        cache_hint = options.get('cache_hint')
        self.use_cache = builder.get_object('remember')
        self.use_cache.set_label(_('Cache password'))
        self.use_cache.set_visible(allow_cache)
        self.use_cache.set_active(cache_hint)

        label = builder.get_object('message')
        label.set_label(message)
        window.set_title(title)
        window.set_keep_above(True)
        super().__init__(window)

    def on_show_password(self, button):
        self.entry.set_visibility(button.get_active())

    async def on_open_keyfile(self, button):
        gtk_dialog = Gtk.FileChooserDialog(
            _("Open a keyfile to unlock the LUKS device"), self.window,
            Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        with Dialog(gtk_dialog) as dialog:
            response = await dialog
            if response == Gtk.ResponseType.OK:
                with open(dialog.window.get_filename(), 'rb') as f:
                    self.content = f.read()
                self.window.response(response)

    def get_text(self):
        if self.content is not None:
            return self.content
        return self.entry.get_text()


async def password_dialog(key, title, message, options):
    """
    Show a Gtk password dialog.

    :returns: the password or ``None`` if the user aborted the operation
    :raises RuntimeError: if Gtk can not be properly initialized
    """
    with PasswordDialog.create(key, title, message, options) as dialog:
        response = await dialog
        if response == Gtk.ResponseType.OK:
            return PasswordResult(dialog.get_text(),
                                  dialog.use_cache.get_active())
        return None


def get_password_gui(device, options):
    """Get the password to unlock a device from GUI."""
    text = _('Enter password for {0.device_presentation}: ', device)
    try:
        return password_dialog(device.id_uuid, 'udiskie', text, options)
    except RuntimeError:
        return None


async def get_password_tty(device, options):
    """Get the password to unlock a device from terminal."""
    # TODO: make this a TRUE async
    text = _('Enter password for {0.device_presentation}: ', device)
    try:
        return PasswordResult(getpass.getpass(text))
    except EOFError:
        print("")
        return None


class DeviceCommand:

    """
    Launcher that starts user-defined password prompts. The command can be
    specified in terms of a command line template.
    """

    def __init__(self, argv, capture=False, **extra):
        """Create the launcher object from the command line template."""
        if isinstance(argv, str):
            self.argv = shlex.split(argv)
        else:
            self.argv = argv
        self.capture = capture
        self.extra = extra.copy()
        # obtain a list of used fields names
        formatter = string.Formatter()
        self.used_attrs = set()
        for arg in self.argv:
            for text, kwd, spec, conv in formatter.parse(arg):
                if kwd is None:
                    continue
                if kwd in DeviceFilter.VALID_PARAMETERS:
                    self.used_attrs.add(kwd)
                if kwd not in DeviceFilter.VALID_PARAMETERS and \
                        kwd not in self.extra:
                    self.extra[kwd] = None
                    logging.getLogger(__name__).error(_(
                        'Unknown device attribute {!r} in format string: {!r}',
                        kwd, arg))

    async def __call__(self, device):
        """
        Invoke the subprocess to ask the user to enter a password for unlocking
        the specified device.
        """
        attrs = {attr: getattr(device, attr) for attr in self.used_attrs}
        attrs.update(self.extra)
        argv = [arg.format(**attrs) for arg in self.argv]
        try:
            stdout = await exec_subprocess(argv, self.capture)
        except subprocess.CalledProcessError:
            return None
        # Remove trailing newline for text answers, but not for binary
        # keyfiles. This logic is a guess that may cause bugs for some users:(
        if stdout and stdout.endswith(b'\n') and is_utf8(stdout):
            stdout = stdout[:-1]
        return stdout

    async def password(self, device, options):
        text = await self(device)
        return PasswordResult(text)


def password(password_command):
    """Create a password prompt function."""
    gui = lambda: has_Gtk() and get_password_gui
    tty = lambda: sys.stdin.isatty() and get_password_tty
    if password_command == 'builtin:gui':
        return gui() or tty()
    elif password_command == 'builtin:tty':
        return tty() or gui()
    elif password_command:
        return DeviceCommand(password_command, capture=True).password
    else:
        return None


def browser(browser_name='xdg-open'):

    """Create a browse-directory function."""

    if not browser_name:
        return None
    argv = shlex.split(browser_name)
    executable = find_executable(argv[0])
    if executable is None:
        # Why not raise an exception? -I think it is more convenient (for
        # end users) to have a reasonable default, without enforcing it.
        logging.getLogger(__name__).warn(
            _("Can't find file browser: {0!r}. "
              "You may want to change the value for the '-f' option.",
              browser_name))
        return None

    def browse(path):
        return subprocess.Popen(argv + [path])

    return browse


def notify_command(command_format, mounter):
    """
    Command notification tool.

    This works similar to Notify, but will issue command instead of showing
    the notifications on the desktop. This can then be used to react to events
    from shell scripts.

    The command can contain modern pythonic format placeholders like:
    {device_file}. The following placeholders are supported:
    event, device_file, device_id, device_size, drive, drive_label, id_label,
    id_type, id_usage, id_uuid, mount_path, root

    :param str command_format: command to run when an event occurs.
    :param mounter: Mounter object
    """
    udisks = mounter.udisks
    for event in ['device_mounted', 'device_unmounted',
                  'device_locked', 'device_unlocked',
                  'device_added', 'device_removed',
                  'job_failed']:
        udisks.connect(event, run_bg(DeviceCommand(
            command_format, event=event, capture=False)))



"""
Make sure that the correct versions of gobject introspection dependencies
are installed.
"""

import os
import logging

from gi import require_version

from .locale import _

require_version('Gio', '2.0')
require_version('GLib', '2.0')


def check_call(exc_type, func, *args):
    try:
        func(*args)
        return True
    except exc_type:
        return False


def check_version(package, version):
    if check_call(ValueError, require_version, package, version):
        return (package, version)


_in_X = bool(os.environ.get('DISPLAY'))
_in_Wayland = bool(os.environ.get('WAYLAND_DISPLAY'))
if not _in_X and not _in_Wayland and os.environ.get('XDG_RUNTIME_DIR'):
    _in_Wayland = os.path.exists(os.path.join(
        os.environ.get('XDG_RUNTIME_DIR'), 'wayland-0'))

_has_Gtk = (3 if check_version('Gtk', '3.0') else
            2 if check_version('Gtk', '2.0') else
            0)

_has_Notify = check_version('Notify', '0.7')
_has_AppIndicator3 = (
    check_version('AyatanaAppIndicator3', '0.1') or
    check_version('AppIndicator3', '0.1')
)


def require_Gtk(min_version=2):
    """
    Make sure Gtk is properly initialized.

    :raises RuntimeError: if Gtk can not be properly initialized
    """
    if not (_in_X or _in_Wayland):
        raise RuntimeError('Not in X or Wayland session.')
    if _has_Gtk < min_version:
        raise RuntimeError('Module gi.repository.Gtk not available!')
    if _has_Gtk == 2:
        logging.getLogger(__name__).warn(
            _("Missing runtime dependency GTK 3. Falling back to GTK 2 "
              "for password prompt"))
    from gi.repository import Gtk
    # if we attempt to create any GUI elements with no X server running the
    # program will just crash, so let's make a way to catch this case:
    if not Gtk.init_check(None)[0]:
        raise RuntimeError(_("X server not connected!"))
    return Gtk


def require_Notify():
    if not _has_Notify:
        raise RuntimeError('Module gi.repository.Notify not available!')
    from gi.repository import Notify
    return Notify


def require_AppIndicator3():
    if _has_AppIndicator3 == ('AppIndicator3', '0.1'):
        from gi.repository import AppIndicator3
    elif _has_AppIndicator3 == ('AyatanaAppIndicator3', '0.1'):
        from gi.repository import AyatanaAppIndicator3 as AppIndicator3
    else:
        raise RuntimeError('Module gi.repository.AppIndicator3 not available!')
    return AppIndicator3


def has_Notify():
    return check_call((RuntimeError, ImportError), require_Notify)


def has_Gtk(min_version=2):
    return check_call((RuntimeError, ImportError), require_Gtk, min_version)


def has_AppIndicator3():
    return check_call((RuntimeError, ImportError), require_AppIndicator3)
